#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Callback registry implementation."""

from typing import Any, Callable, Dict, List, Tuple

from typing_extensions import ParamSpec

from zenml.logger import get_logger

P = ParamSpec("P")

logger = get_logger(__name__)


class CallbackRegistry:
    """Callback registry class."""

    def __init__(self) -> None:
        """Initializes the callback registry."""
        self._callbacks: List[
            Tuple[Callable[P, Any], Tuple[Any], Dict[str, Any]]
        ] = []

    def register_callback(
        self, callback: Callable[P, Any], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        """Register a callback.

        Args:
            callback: The callback to register.
            *args: Arguments to call the callback with.
            **kwargs: Keyword arguments to call the callback with.
        """
        self._callbacks.append((callback, args, kwargs))  # type: ignore[arg-type]

    def reset(self) -> None:
        """Reset the callbacks."""
        self._callbacks = []

    def execute_callbacks(self, raise_on_exception: bool) -> None:
        """Execute all registered callbacks.

        Args:
            raise_on_exception: If True, exceptions raised during the execution
                of the callbacks will be raised. If False, a warning with the
                exception will be logged instead.

        Raises:
            Exception: Exceptions raised in any of the callbacks if
                `raise_on_exception` is set to True.
        """
        for callback, args, kwargs in self._callbacks:
            try:
                callback(*args, **kwargs)
            except Exception as e:
                if raise_on_exception:
                    raise e
                else:
                    logger.warning("Failed to run callback: %s", str(e))



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for flavors."""

from uuid import UUID

from fastapi import APIRouter, Depends, Security

from zenml.constants import API, FLAVORS, VERSION_1
from zenml.models import (
    FlavorFilter,
    FlavorRequest,
    FlavorResponse,
    FlavorUpdate,
    Page,
)
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import Action, ResourceType
from zenml.zen_server.rbac.utils import verify_permission
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + FLAVORS,
    tags=["flavors"],
    responses={401: error_response, 403: error_response},
)


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def list_flavors(
    flavor_filter_model: FlavorFilter = Depends(make_dependable(FlavorFilter)),
    hydrate: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[FlavorResponse]:
    """Returns all flavors.

    Args:
        flavor_filter_model: Filter model used for pagination, sorting,
                             filtering
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        All flavors.
    """
    return verify_permissions_and_list_entities(
        filter_model=flavor_filter_model,
        resource_type=ResourceType.FLAVOR,
        list_method=zen_store().list_flavors,
        hydrate=hydrate,
    )


@router.get(
    "/{flavor_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def get_flavor(
    flavor_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> FlavorResponse:
    """Returns the requested flavor.

    Args:
        flavor_id: ID of the flavor.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        The requested stack.
    """
    return verify_permissions_and_get_entity(
        id=flavor_id, get_method=zen_store().get_flavor, hydrate=hydrate
    )


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def create_flavor(
    flavor: FlavorRequest,
    _: AuthContext = Security(authorize),
) -> FlavorResponse:
    """Creates a stack component flavor.

    Args:
        flavor: Stack component flavor to register.

    Returns:
        The created stack component flavor.
    """
    return verify_permissions_and_create_entity(
        request_model=flavor,
        create_method=zen_store().create_flavor,
    )


@router.put(
    "/{flavor_id}",
    responses={401: error_response, 409: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def update_flavor(
    flavor_id: UUID,
    flavor_update: FlavorUpdate,
    _: AuthContext = Security(authorize),
) -> FlavorResponse:
    """Updates a flavor.

    # noqa: DAR401

    Args:
        flavor_id: ID of the flavor to update.
        flavor_update: Flavor update.

    Returns:
        The updated flavor.
    """
    return verify_permissions_and_update_entity(
        id=flavor_id,
        update_model=flavor_update,
        get_method=zen_store().get_flavor,
        update_method=zen_store().update_flavor,
    )


@router.delete(
    "/{flavor_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_flavor(
    flavor_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Deletes a flavor.

    Args:
        flavor_id: ID of the flavor.
    """
    verify_permissions_and_delete_entity(
        id=flavor_id,
        get_method=zen_store().get_flavor,
        delete_method=zen_store().delete_flavor,
    )


@router.patch(
    "/sync",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def sync_flavors(
    _: AuthContext = Security(authorize),
) -> None:
    """Purge all in-built and integration flavors from the DB and sync.

    Returns:
        None if successful. Raises an exception otherwise.
    """
    verify_permission(resource_type=ResourceType.FLAVOR, action=Action.UPDATE)
    return zen_store()._sync_flavors()



"""Release [0.57.0].

Revision ID: 0.57.0
Revises: 0.57.0rc2
Create Date: 2024-05-02 13:05:56.733082

"""

# revision identifiers, used by Alembic.
revision = "0.57.0"
down_revision = "0.57.0rc2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML server deployer singleton implementation."""

from typing import ClassVar, Dict, Generator, Optional, Type

from zenml.config.global_config import GlobalConfiguration
from zenml.enums import ServerProviderType, StoreType
from zenml.logger import get_logger
from zenml.utils.singleton import SingletonMetaClass
from zenml.zen_server.deploy.base_provider import BaseServerProvider
from zenml.zen_server.deploy.deployment import (
    LocalServerDeployment,
    LocalServerDeploymentConfig,
)
from zenml.zen_server.deploy.exceptions import (
    ServerDeploymentError,
    ServerDeploymentNotFoundError,
    ServerProviderNotFoundError,
)
from zenml.zen_stores.base_zen_store import BaseZenStore
from zenml.zen_stores.rest_zen_store import RestZenStoreConfiguration

logger = get_logger(__name__)


class LocalServerDeployer(metaclass=SingletonMetaClass):
    """Local server deployer singleton.

    This class is responsible for managing the various server provider
    implementations and for directing server deployment lifecycle requests to
    the responsible provider. It acts as a facade built on top of the various
    server providers.
    """

    _providers: ClassVar[Dict[ServerProviderType, BaseServerProvider]] = {}

    @classmethod
    def register_provider(cls, provider: Type[BaseServerProvider]) -> None:
        """Register a server provider.

        Args:
            provider: The server provider to register.

        Raises:
            TypeError: If a provider with the same type is already registered.
        """
        if provider.TYPE in cls._providers:
            raise TypeError(
                f"Server provider '{provider.TYPE}' is already registered."
            )
        logger.debug(f"Registering server provider '{provider.TYPE}'.")
        cls._providers[provider.TYPE] = provider()

    @classmethod
    def get_provider(
        cls, provider_type: ServerProviderType
    ) -> BaseServerProvider:
        """Get the server provider associated with a provider type.

        Args:
            provider_type: The server provider type.

        Returns:
            The server provider associated with the provider type.

        Raises:
            ServerProviderNotFoundError: If no provider is registered for the
                given provider type.
        """
        if provider_type not in cls._providers:
            raise ServerProviderNotFoundError(
                f"Server provider '{provider_type}' is not registered."
            )
        return cls._providers[provider_type]

    def initialize_local_database(self) -> None:
        """Initialize the local ZenML database."""
        default_store_cfg = GlobalConfiguration().get_default_store()
        BaseZenStore.create_store(default_store_cfg)

    def deploy_server(
        self,
        config: LocalServerDeploymentConfig,
        timeout: Optional[int] = None,
        restart: bool = False,
    ) -> LocalServerDeployment:
        """Deploy the local ZenML server or update the existing deployment.

        Args:
            config: The server deployment configuration.
            timeout: The timeout in seconds to wait until the deployment is
                successful. If not supplied, the default timeout value specified
                by the provider is used.
            restart: If True, the existing server deployment will be torn down
                and a new server will be deployed.

        Returns:
            The local server deployment.
        """
        # Ensure that the local database is always initialized before any local
        # server is deployed or updated.
        self.initialize_local_database()

        try:
            self.get_server()
        except ServerDeploymentNotFoundError:
            pass
        else:
            return self.update_server(
                config=config, timeout=timeout, restart=restart
            )

        provider_name = config.provider.value
        provider = self.get_provider(config.provider)

        logger.info(f"Deploying a local {provider_name} ZenML server.")
        return provider.deploy_server(config, timeout=timeout)

    def update_server(
        self,
        config: LocalServerDeploymentConfig,
        timeout: Optional[int] = None,
        restart: bool = False,
    ) -> LocalServerDeployment:
        """Update an existing local ZenML server deployment.

        Args:
            config: The new server deployment configuration.
            timeout: The timeout in seconds to wait until the deployment is
                successful. If not supplied, a default timeout value of 30
                seconds is used.
            restart: If True, the existing server deployment will be torn down
                and a new server will be deployed.

        Returns:
            The updated server deployment.
        """
        # this will also raise ServerDeploymentNotFoundError if the server
        # does not exist
        existing_server = self.get_server()

        provider = self.get_provider(config.provider)
        if existing_server.config.provider != config.provider or restart:
            existing_provider = self.get_provider(
                existing_server.config.provider
            )

            # Tear down the existing server deployment
            existing_provider.remove_server(
                existing_server.config, timeout=timeout
            )

            # Deploy a new server with the new provider
            return provider.deploy_server(config, timeout=timeout)

        return provider.update_server(config, timeout=timeout)

    def remove_server(
        self,
        timeout: Optional[int] = None,
    ) -> None:
        """Tears down and removes all resources and files associated with the local ZenML server deployment.

        Args:
            timeout: The timeout in seconds to wait until the deployment is
                successfully torn down. If not supplied, a provider specific
                default timeout value is used.
        """
        # this will also raise ServerDeploymentNotFoundError if the server
        # does not exist
        try:
            server = self.get_server()
        except ServerDeploymentNotFoundError:
            return

        provider_name = server.config.provider.value
        provider = self.get_provider(server.config.provider)

        if self.is_connected_to_server():
            try:
                self.disconnect_from_server()
            except Exception as e:
                logger.warning(
                    f"Failed to disconnect from the local server: {e}"
                )

        logger.info(f"Tearing down the local {provider_name} ZenML server.")
        provider.remove_server(server.config, timeout=timeout)

    def is_connected_to_server(self) -> bool:
        """Check if the ZenML client is currently connected to the local ZenML server.

        Returns:
            True if the ZenML client is connected to the local ZenML server, False
            otherwise.
        """
        # this will also raise ServerDeploymentNotFoundError if the server
        # does not exist
        try:
            server = self.get_server()
        except ServerDeploymentNotFoundError:
            return False

        gc = GlobalConfiguration()
        return (
            server.status is not None
            and server.status.url is not None
            and gc.store_configuration.url == server.status.url
        )

    def connect_to_server(
        self,
    ) -> None:
        """Connect to the local ZenML server instance.

        Raises:
            ServerDeploymentError: If the local ZenML server is not running or
                is unreachable.
        """
        # this will also raise ServerDeploymentNotFoundError if the server
        # does not exist
        server = self.get_server()
        provider_name = server.config.provider.value

        gc = GlobalConfiguration()
        if not server.status or not server.status.url:
            raise ServerDeploymentError(
                f"The local {provider_name} ZenML server is not currently "
                "running or is unreachable."
            )

        store_config = RestZenStoreConfiguration(
            url=server.status.url,
        )

        if gc.store_configuration == store_config:
            logger.info(
                "Your client is already connected to the local "
                f"{provider_name} ZenML server."
            )
            return

        logger.info(
            f"Connecting to the local {provider_name} ZenML server "
            f"({store_config.url})."
        )

        gc.set_store(store_config)

        logger.info(
            f"Connected to the local {provider_name} ZenML server "
            f"({store_config.url})."
        )

    def disconnect_from_server(
        self,
    ) -> None:
        """Disconnect from the ZenML server instance."""
        gc = GlobalConfiguration()
        store_cfg = gc.store_configuration

        if store_cfg.type != StoreType.REST:
            logger.info(
                "Your client is not currently connected to a ZenML server."
            )
            return

        logger.info(
            f"Disconnecting from the local ({store_cfg.url}) ZenML server."
        )

        gc.set_default_store()

        logger.info("Disconnected from the local ZenML server.")

    def get_server(
        self,
    ) -> LocalServerDeployment:
        """Get the local server deployment.

        Returns:
            The local server deployment.

        Raises:
            ServerDeploymentNotFoundError: If no local server deployment is
                found.
        """
        for provider in self._providers.values():
            try:
                return provider.get_server(
                    LocalServerDeploymentConfig(provider=provider.TYPE)
                )
            except ServerDeploymentNotFoundError:
                pass

        raise ServerDeploymentNotFoundError(
            "No local server deployment was found."
        )

    def get_server_logs(
        self,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Retrieve the logs for the local ZenML server.

        Args:
            follow: if True, the logs will be streamed as they are written
            tail: only retrieve the last NUM lines of log output.

        Returns:
            A generator that can be accessed to get the service logs.
        """
        # this will also raise ServerDeploymentNotFoundError if the server
        # does not exist
        server = self.get_server()

        provider_name = server.config.provider.value
        provider = self.get_provider(server.config.provider)

        logger.info(
            f"Fetching logs from the local {provider_name} ZenML server..."
        )
        return provider.get_server_logs(
            server.config, follow=follow, tail=tail
        )



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Code utilities."""

import hashlib
import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import IO, TYPE_CHECKING, Dict, Optional

from zenml.client import Client
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.utils import io_utils, source_utils, string_utils
from zenml.utils.archivable import Archivable, ArchiveType

if TYPE_CHECKING:
    from git.repo.base import Repo

    from zenml.artifact_stores import BaseArtifactStore


logger = get_logger(__name__)


class CodeArchive(Archivable):
    """Code archive class.

    This class is used to archive user code before uploading it to the artifact
    store. If the user code is stored in a Git repository, only files not
    excluded by gitignores will be included in the archive.
    """

    def __init__(self, root: Optional[str] = None) -> None:
        """Initialize the object.

        Args:
            root: Root directory of the archive.
        """
        super().__init__()
        self._root = root

    @property
    def git_repo(self) -> Optional["Repo"]:
        """Git repository active at the code archive root.

        Returns:
            The git repository if available.
        """
        try:
            # These imports fail when git is not installed on the machine
            from git.exc import InvalidGitRepositoryError
            from git.repo.base import Repo
        except ImportError:
            return None

        try:
            git_repo = Repo(path=self._root, search_parent_directories=True)
        except InvalidGitRepositoryError:
            return None

        return git_repo

    def _get_all_files(self, archive_root: str) -> Dict[str, str]:
        """Get all files inside the archive root.

        Args:
            archive_root: The root directory from which to get all files.

        Returns:
            All files inside the archive root.
        """
        all_files = {}
        for root, _, files in os.walk(archive_root):
            for file in files:
                file_path = os.path.join(root, file)
                path_in_archive = os.path.relpath(file_path, archive_root)
                all_files[path_in_archive] = file_path

        return all_files

    def get_files(self) -> Dict[str, str]:
        """Gets all regular files that should be included in the archive.

        Raises:
            RuntimeError: If the code archive would not include any files.

        Returns:
            A dict {path_in_archive: path_on_filesystem} for all regular files
            in the archive.
        """
        if not self._root:
            return {}

        all_files = {}

        if repo := self.git_repo:
            try:
                result = repo.git.ls_files(
                    "--cached",
                    "--others",
                    "--modified",
                    "--exclude-standard",
                    self._root,
                )
            except Exception as e:
                logger.warning(
                    "Failed to get non-ignored files from git: %s", str(e)
                )
                all_files = self._get_all_files(archive_root=self._root)
            else:
                for file in result.split():
                    file_path = os.path.join(repo.working_dir, file)
                    path_in_archive = os.path.relpath(file_path, self._root)

                    if os.path.exists(file_path):
                        all_files[path_in_archive] = file_path
        else:
            all_files = self._get_all_files(archive_root=self._root)

        if not all_files:
            raise RuntimeError(
                "The code archive to be uploaded does not contain any files. "
                "This is probably because all files in your source root "
                f"`{self._root}` are ignored by a .gitignore file."
            )

        # Explicitly remove .zen directories as we write an updated version
        # to disk everytime ZenML is called. This updates the mtime of the
        # file, which invalidates the code upload caching. The values in
        # the .zen directory are not needed anyway as we set them as
        # environment variables.
        all_files = {
            path_in_archive: file_path
            for path_in_archive, file_path in sorted(all_files.items())
            if ".zen" not in Path(path_in_archive).parts[:-1]
        }

        return all_files

    def write_archive(
        self,
        output_file: IO[bytes],
        archive_type: ArchiveType = ArchiveType.TAR_GZ,
    ) -> None:
        """Writes an archive of the build context to the given file.

        Args:
            output_file: The file to write the archive to.
            archive_type: The type of archive to create.
        """
        super().write_archive(
            output_file=output_file, archive_type=archive_type
        )
        archive_size = os.path.getsize(output_file.name)
        if archive_size > 20 * 1024 * 1024:
            logger.warning(
                "Code archive size: `%s`. If you believe this is "
                "unreasonably large, make sure to version your code in git and "
                "ignore unnecessary files using a `.gitignore` file.",
                string_utils.get_human_readable_filesize(archive_size),
            )


def compute_file_hash(file: IO[bytes]) -> str:
    """Compute a hash of the content of a file.

    This function will not seek the file before or after the hash computation.
    This means that the content will be computed based on the current cursor
    until the end of the file.

    Args:
        file: The file for which to compute the hash.

    Returns:
        A hash of the file content.
    """
    hash_ = hashlib.sha1()  # nosec

    while True:
        data = file.read(64 * 1024)
        if not data:
            break
        hash_.update(data)

    return hash_.hexdigest()


def upload_code_if_necessary(code_archive: CodeArchive) -> str:
    """Upload code to the artifact store if necessary.

    This function computes a hash of the code to be uploaded, and if an archive
    with the same hash already exists it will not re-upload but instead return
    the path to the existing archive.

    Args:
        code_archive: The code archive to upload.

    Returns:
        The path where the archived code is uploaded.
    """
    artifact_store = Client().active_stack.artifact_store

    with tempfile.NamedTemporaryFile(
        mode="w+b", delete=False, suffix=".tar.gz"
    ) as f:
        code_archive.write_archive(f)
        archive_path = f.name
        archive_hash = compute_file_hash(f)

    upload_dir = os.path.join(artifact_store.path, "code_uploads")
    fileio.makedirs(upload_dir)
    upload_path = os.path.join(upload_dir, f"{archive_hash}.tar.gz")
    upload_path = io_utils.sanitize_remote_path(upload_path)

    if not fileio.exists(upload_path):
        archive_size = string_utils.get_human_readable_filesize(
            os.path.getsize(archive_path)
        )
        logger.info(
            "Uploading code to `%s` (Size: %s).", upload_path, archive_size
        )
        fileio.copy(archive_path, upload_path)
        logger.info("Code upload finished.")
    else:
        logger.info("Code already exists in artifact store, skipping upload.")

    if os.path.exists(archive_path):
        os.remove(archive_path)

    return upload_path


def download_and_extract_code(code_path: str, extract_dir: str) -> None:
    """Download and extract code.

    Args:
        code_path: Path where the code is uploaded.
        extract_dir: Directory where to code should be extracted to.
    """
    download_path = os.path.basename(code_path)
    fileio.copy(code_path, download_path)

    shutil.unpack_archive(filename=download_path, extract_dir=extract_dir)
    os.remove(download_path)


def download_code_from_artifact_store(
    code_path: str, artifact_store: "BaseArtifactStore"
) -> None:
    """Download code from the artifact store.

    Args:
        code_path: Path where the code is stored.
        artifact_store: The artifact store to use for the download.

    Raises:
        RuntimeError: If the code is stored in an artifact store which is
            not active.
    """
    logger.info("Downloading code from artifact store path `%s`.", code_path)

    if not code_path.startswith(artifact_store.path):
        raise RuntimeError(
            "The code is not stored in the artifact store "
            f"{artifact_store.name} that was passed to download it."
        )

    # Make sure we register the artifact store filesystem here so the
    # fileio.copy call will pick up the right credentials
    artifact_store._register()

    extract_dir = os.path.abspath("code")
    os.makedirs(extract_dir, exist_ok=True)

    download_and_extract_code(code_path=code_path, extract_dir=extract_dir)

    source_utils.set_custom_source_root(extract_dir)
    sys.path.insert(0, extract_dir)
    os.chdir(extract_dir)


def _get_notebook_upload_dir(artifact_store: "BaseArtifactStore") -> str:
    """Get the upload directory for code extracted from notebook cells.

    Args:
        artifact_store: The artifact store in which the directory should be.

    Returns:
        The upload directory for code extracted from notebook cells.
    """
    return os.path.join(artifact_store.path, "notebook_code")


def upload_notebook_code(
    artifact_store: "BaseArtifactStore", cell_code: str, file_name: str
) -> None:
    """Upload code extracted from a notebook cell.

    Args:
        artifact_store: The artifact store in which to upload the code.
        cell_code: The notebook cell code.
        file_name: The filename to use for storing the cell code.
    """
    upload_dir = _get_notebook_upload_dir(artifact_store=artifact_store)
    fileio.makedirs(upload_dir)
    upload_path = os.path.join(upload_dir, file_name)

    if not fileio.exists(upload_path):
        with fileio.open(upload_path, "wb") as f:
            f.write(cell_code.encode())

        logger.info("Uploaded notebook cell code to %s.", upload_path)


def download_notebook_code(
    artifact_store: "BaseArtifactStore", file_name: str, download_path: str
) -> None:
    """Download code extracted from a notebook cell.

    Args:
        artifact_store: The artifact store from which to download the code.
        file_name: The name of the code file.
        download_path: The local path where the file should be downloaded to.

    Raises:
        FileNotFoundError: If no file with the given filename exists in this
            artifact store.
    """
    code_dir = _get_notebook_upload_dir(artifact_store=artifact_store)
    code_path = os.path.join(code_dir, file_name)

    if not fileio.exists(code_path):
        raise FileNotFoundError(
            f"Notebook code at path {code_path} not found."
        )

    fileio.copy(code_path, download_path)



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utils for schemas."""

import json
import math
from typing import Any, Dict, List, Type, TypeVar, cast

from sqlalchemy.orm import InstrumentedAttribute
from sqlmodel import Relationship

from zenml.metadata.metadata_types import MetadataType
from zenml.models import BaseResponse, Page, RunMetadataEntry
from zenml.zen_stores.schemas.base_schemas import BaseSchema

S = TypeVar("S", bound=BaseSchema)


def jl_arg(column: Any) -> InstrumentedAttribute[Any]:
    """Cast a SQLModel column to a joinedload argument.

    Args:
        column: The column.

    Returns:
        The column cast to a joinedload argument.
    """
    return cast(InstrumentedAttribute[Any], column)


def get_page_from_list(
    items_list: List[S],
    response_model: Type[BaseResponse],  # type: ignore[type-arg]
    size: int = 5,
    page: int = 1,
    include_resources: bool = False,
    include_metadata: bool = False,
) -> Page[BaseResponse]:  # type: ignore[type-arg]
    """Converts list of schemas into page of response models.

    Args:
        items_list: List of schemas
        response_model: Response model
        size: Page size
        page: Page number
        include_metadata: Whether metadata should be included in response models
        include_resources: Whether resources should be included in response models

    Returns:
        A page of list items.
    """
    total = len(items_list)
    if total == 0:
        total_pages = 1
    else:
        total_pages = math.ceil(total / size)

    start = (page - 1) * size
    end = start + size

    page_items = [
        item.to_model(
            include_metadata=include_metadata,
            include_resources=include_resources,
        )
        for item in items_list[start:end]
    ]
    return Page[response_model](  # type: ignore[valid-type]
        index=page,
        max_size=size,
        total_pages=total_pages,
        total=total,
        items=page_items,
    )


class RunMetadataInterface:
    """The interface for entities with run metadata."""

    run_metadata = Relationship()

    def fetch_metadata_collection(
        self, **kwargs: Any
    ) -> Dict[str, List[RunMetadataEntry]]:
        """Fetches all the metadata entries related to the entity.

        Args:
            **kwargs: Keyword arguments.

        Returns:
            A dictionary, where the key is the key of the metadata entry
                and the values represent the list of entries with this key.
        """
        metadata_collection: Dict[str, List[RunMetadataEntry]] = {}

        for rm in self.run_metadata:
            if rm.key not in metadata_collection:
                metadata_collection[rm.key] = []
            metadata_collection[rm.key].append(
                RunMetadataEntry(
                    value=json.loads(rm.value),
                    created=rm.created,
                )
            )

        return metadata_collection

    def fetch_metadata(self, **kwargs: Any) -> Dict[str, MetadataType]:
        """Fetches the latest metadata entry related to the entity.

        Args:
            **kwargs: Keyword arguments to pass to the metadata collection.

        Returns:
            A dictionary, where the key is the key of the metadata entry
                and the values represent the latest entry with this key.
        """
        metadata_collection = self.fetch_metadata_collection(**kwargs)
        return {
            k: sorted(v, key=lambda x: x.created, reverse=True)[0].value
            for k, v in metadata_collection.items()
        }


def get_resource_type_name(schema_class: Type[BaseSchema]) -> str:
    """Get the name of a resource from a schema class.

    Args:
        schema_class: The schema class to get the name of.

    Returns:
        The name of the resource.
    """
    entity_name = schema_class.__tablename__
    assert isinstance(entity_name, str)
    # Some entities are plural, some are singular, some have multiple words
    # in their table name connected by underscores (e.g. pipeline_run)
    return entity_name.replace("_", " ").rstrip("s")



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation for the BentoML local deployment service."""

import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from bentoml import AsyncHTTPClient, SyncHTTPClient
from pydantic import BaseModel, Field

from zenml.constants import DEFAULT_LOCAL_SERVICE_IP_ADDRESS
from zenml.integrations.bentoml.constants import (
    BENTOML_DEFAULT_PORT,
    BENTOML_HEALTHCHECK_URL_PATH,
    BENTOML_PREDICTION_URL_PATH,
)
from zenml.logger import get_logger
from zenml.models.v2.misc.service import ServiceType
from zenml.services import (
    HTTPEndpointHealthMonitor,
    HTTPEndpointHealthMonitorConfig,
    LocalDaemonService,
    LocalDaemonServiceConfig,
    LocalDaemonServiceEndpoint,
    LocalDaemonServiceEndpointConfig,
    ServiceEndpointProtocol,
)
from zenml.services.service import BaseDeploymentService

if TYPE_CHECKING:
    from zenml.integrations.bentoml.model_deployers.bentoml_model_deployer import (  # noqa
        BentoMLModelDeployer,
    )

logger = get_logger(__name__)

BENTOML_LOCAL_DEPLOYMENT_SERVICE_NAME = "bentoml-local-deployment"


class BentoMLDeploymentEndpointConfig(LocalDaemonServiceEndpointConfig):
    """BentoML deployment service configuration.

    Attributes:
        prediction_url_path: URI subpath for prediction requests
    """

    prediction_url_path: str


class BentoMLDeploymentEndpoint(LocalDaemonServiceEndpoint):
    """A service endpoint exposed by the BentoML deployment daemon.

    Attributes:
        config: service endpoint configuration
    """

    config: BentoMLDeploymentEndpointConfig
    monitor: HTTPEndpointHealthMonitor

    @property
    def prediction_url(self) -> Optional[str]:
        """Gets the prediction URL for the endpoint.

        Returns:
            the prediction URL for the endpoint
        """
        uri = self.status.uri
        if not uri:
            return None
        return os.path.join(uri, self.config.prediction_url_path)


class SSLBentoMLParametersConfig(BaseModel):
    """BentoML SSL parameters configuration.

    Attributes:
        ssl_certfile: SSL certificate file
        ssl_keyfile: SSL key file
        ssl_keyfile_password: SSL key file password
        ssl_version: SSL version
        ssl_cert_reqs: SSL certificate requirements
        ssl_ca_certs: SSL CA certificates
        ssl_ciphers: SSL ciphers
    """

    ssl_certfile: Optional[str] = None
    ssl_keyfile: Optional[str] = None
    ssl_keyfile_password: Optional[str] = None
    ssl_version: Optional[int] = None
    ssl_cert_reqs: Optional[int] = None
    ssl_ca_certs: Optional[str] = None
    ssl_ciphers: Optional[str] = None


class BentoMLLocalDeploymentConfig(LocalDaemonServiceConfig):
    """BentoML model deployment configuration.

    Attributes:
        model_name: name of the model to deploy
        model_uri: URI of the model to deploy
        port: port to expose the service on
        bento_tag: Bento package to deploy. A bento tag is a combination of the
            name of the bento and its version.
        workers: number of workers to use
        backlog: number of requests to queue
        production: whether to run in production mode
        working_dir: working directory for the service
        host: host to expose the service on
        ssl_parameters: SSL parameters for the Bentoml deployment
    """

    model_name: str
    model_uri: str
    bento_tag: str
    bento_uri: Optional[str] = None
    apis: List[str] = []
    workers: int = 1
    port: Optional[int] = None
    backlog: int = 2048
    production: bool = False
    working_dir: str
    host: Optional[str] = None
    ssl_parameters: Optional[SSLBentoMLParametersConfig] = Field(
        default_factory=SSLBentoMLParametersConfig
    )


class BentoMLLocalDeploymentService(LocalDaemonService, BaseDeploymentService):
    """BentoML deployment service used to start a local prediction server for BentoML models.

    Attributes:
        SERVICE_TYPE: a service type descriptor with information describing
            the BentoML deployment service class
        config: service configuration
        endpoint: optional service endpoint
    """

    SERVICE_TYPE = ServiceType(
        name=BENTOML_LOCAL_DEPLOYMENT_SERVICE_NAME,
        type="model-serving",
        flavor="bentoml",
        description="BentoML local prediction service",
        logo_url="https://public-flavor-logos.s3.eu-central-1.amazonaws.com/model_deployer/bentoml.png",
    )

    config: BentoMLLocalDeploymentConfig
    endpoint: BentoMLDeploymentEndpoint

    def __init__(
        self,
        config: Union[BentoMLLocalDeploymentConfig, Dict[str, Any]],
        **attrs: Any,
    ) -> None:
        """Initialize the BentoML deployment service.

        Args:
            config: service configuration
            attrs: additional attributes to set on the service
        """
        # ensure that the endpoint is created before the service is initialized
        # TODO [ENG-700]: implement a service factory or builder for BentoML
        #   deployment services
        if (
            isinstance(config, BentoMLLocalDeploymentConfig)
            and "endpoint" not in attrs
        ):
            endpoint = BentoMLDeploymentEndpoint(
                config=BentoMLDeploymentEndpointConfig(
                    protocol=ServiceEndpointProtocol.HTTP,
                    port=config.port
                    if config.port is not None
                    else BENTOML_DEFAULT_PORT,
                    ip_address=config.host or DEFAULT_LOCAL_SERVICE_IP_ADDRESS,
                    prediction_url_path=BENTOML_PREDICTION_URL_PATH,
                ),
                monitor=HTTPEndpointHealthMonitor(
                    config=HTTPEndpointHealthMonitorConfig(
                        healthcheck_uri_path=BENTOML_HEALTHCHECK_URL_PATH,
                    )
                ),
            )
            attrs["endpoint"] = endpoint
        super().__init__(config=config, **attrs)

    def run(self) -> None:
        """Start the service."""
        from bentoml import Service
        from bentoml._internal.service.loader import load

        logger.info(
            "Starting BentoML prediction service as blocking "
            "process... press CTRL+C once to stop it."
        )

        self.endpoint.prepare_for_start()
        ssl_params = self.config.ssl_parameters or SSLBentoMLParametersConfig()
        # verify if to deploy in production mode or development mode
        logger.info("Running in production mode.")
        svc = load(
            bento_identifier=self.config.bento_tag,
            working_dir=self.config.working_dir or ".",
        )

        if isinstance(svc, Service):
            # bentoml<1.2
            from bentoml.serving import serve_http_production

            try:
                serve_http_production(
                    self.config.bento_tag,
                    working_dir=self.config.working_dir,
                    port=self.config.port,
                    api_workers=self.config.workers,
                    host=self.config.host or DEFAULT_LOCAL_SERVICE_IP_ADDRESS,
                    backlog=self.config.backlog,
                    ssl_certfile=ssl_params.ssl_certfile,
                    ssl_keyfile=ssl_params.ssl_keyfile,
                    ssl_keyfile_password=ssl_params.ssl_keyfile_password,
                    ssl_version=ssl_params.ssl_version,
                    ssl_cert_reqs=ssl_params.ssl_cert_reqs,
                    ssl_ca_certs=ssl_params.ssl_ca_certs,
                    ssl_ciphers=ssl_params.ssl_ciphers,
                )
            except KeyboardInterrupt:
                logger.info("Stopping BentoML prediction service...")
        else:
            # bentoml>=1.2
            from _bentoml_impl.server import serve_http  # type: ignore

            svc.inject_config()
            try:
                serve_http(
                    self.config.bento_tag,
                    working_dir=self.config.working_dir or ".",
                    host=self.config.host or DEFAULT_LOCAL_SERVICE_IP_ADDRESS,
                    port=self.config.port,
                    backlog=self.config.backlog,
                    ssl_certfile=ssl_params.ssl_certfile,
                    ssl_keyfile=ssl_params.ssl_keyfile,
                    ssl_keyfile_password=ssl_params.ssl_keyfile_password,
                    ssl_version=ssl_params.ssl_version,
                    ssl_cert_reqs=ssl_params.ssl_cert_reqs,
                    ssl_ca_certs=ssl_params.ssl_ca_certs,
                    ssl_ciphers=ssl_params.ssl_ciphers,
                )
            except KeyboardInterrupt:
                logger.info("Stopping BentoML prediction service...")

    @property
    def prediction_url(self) -> Optional[str]:
        """Get the URI where the http server is running.

        Returns:
            The URI where the http service can be accessed to get more information
            about the service and to make predictions.
        """
        if not self.is_running:
            return None
        return self.endpoint.prediction_url

    @property
    def prediction_apis_urls(self) -> Optional[List[str]]:
        """Get the URI where the prediction api services is answering requests.

        Returns:
            The URI where the prediction service apis can be contacted to process
            HTTP/REST inference requests, or None, if the service isn't running.
        """
        if not self.is_running:
            return None

        if self.config.apis:
            return [
                f"{self.endpoint.prediction_url}/{api}"
                for api in self.config.apis
            ]
        return None

    def predict(
        self, api_endpoint: str, data: "Any", sync: bool = True
    ) -> "Any":
        """Make a prediction using the service.

        Args:
            data: data to make a prediction on
            api_endpoint: the api endpoint to make the prediction on
            sync: if set to False, the prediction will be made asynchronously

        Returns:
            The prediction result.

        Raises:
            Exception: if the service is not running
            ValueError: if the prediction endpoint is unknown.
        """
        if not self.is_running:
            raise Exception(
                "BentoML prediction service is not running. "
                "Please start the service before making predictions."
            )
        if self.endpoint.prediction_url is None:
            raise ValueError("No endpoint known for prediction.")
        if sync:
            client = SyncHTTPClient(self.endpoint.prediction_url)
        else:
            client = AsyncHTTPClient(self.endpoint.prediction_url)
        result = client.call(api_endpoint, data)
        return result



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base orchestrator class."""

import time
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

from pydantic import model_validator

from zenml.constants import (
    ENV_ZENML_PREVENT_CLIENT_SIDE_CACHING,
    handle_bool_env_var,
)
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.exceptions import RunMonitoringError, RunStoppedException
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType
from zenml.orchestrators.publish_utils import (
    publish_pipeline_run_metadata,
    publish_pipeline_run_status_update,
    publish_schedule_metadata,
)
from zenml.orchestrators.step_launcher import StepLauncher
from zenml.orchestrators.utils import get_config_environment_vars
from zenml.stack import Flavor, Stack, StackComponent, StackComponentConfig
from zenml.utils.pydantic_utils import before_validator_handler

if TYPE_CHECKING:
    from zenml.config.step_configurations import Step
    from zenml.models import (
        PipelineDeploymentResponse,
        PipelineRunResponse,
        ScheduleResponse,
        ScheduleUpdate,
    )

logger = get_logger(__name__)


class SubmissionResult:
    """Result of submitting a pipeline run."""

    def __init__(
        self,
        wait_for_completion: Optional[Callable[[], None]] = None,
        metadata: Optional[Dict[str, MetadataType]] = None,
    ):
        """Initialize a submission result.

        Args:
            wait_for_completion: A function that waits for the pipeline run to
                complete. If provided, this will be called after the pipeline
                run was submitted successfully.
            metadata: Metadata for the pipeline run or schedule.
        """
        self.wait_for_completion = wait_for_completion
        self.metadata = metadata


class BaseOrchestratorConfig(StackComponentConfig):
    """Base orchestrator config."""

    @model_validator(mode="before")
    @classmethod
    @before_validator_handler
    def _deprecations(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        """Validate and/or remove deprecated fields.

        Args:
            data: The values to validate.

        Returns:
            The validated values.
        """
        if "custom_docker_base_image_name" in data:
            image_name = data.pop("custom_docker_base_image_name", None)
            if image_name:
                logger.warning(
                    "The 'custom_docker_base_image_name' field has been "
                    "deprecated. To use a custom base container image with your "
                    "orchestrators, please use the DockerSettings in your "
                    "pipeline (see https://docs.zenml.io/concepts/containerization)."
                )

        return data

    @property
    def is_synchronous(self) -> bool:
        """Whether the orchestrator runs synchronous or not.

        Returns:
            Whether the orchestrator runs synchronous or not.
        """
        return False

    @property
    def is_schedulable(self) -> bool:
        """Whether the orchestrator is schedulable or not.

        Returns:
            Whether the orchestrator is schedulable or not.
        """
        return False

    @property
    def supports_client_side_caching(self) -> bool:
        """Whether the orchestrator supports client side caching.

        Returns:
            Whether the orchestrator supports client side caching.
        """
        return True

    @property
    def handles_step_retries(self) -> bool:
        """Whether the orchestrator handles step retries.

        Returns:
            Whether the orchestrator handles step retries.
        """
        return False


class BaseOrchestrator(StackComponent, ABC):
    """Base class for all orchestrators."""

    _active_deployment: Optional["PipelineDeploymentResponse"] = None

    @property
    def config(self) -> BaseOrchestratorConfig:
        """Returns the `BaseOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(BaseOrchestratorConfig, self._config)

    @abstractmethod
    def get_orchestrator_run_id(self) -> str:
        """Returns the run id of the active orchestrator run.

        Important: This needs to be a unique ID and return the same value for
        all steps of a pipeline run.

        Returns:
            The orchestrator run id.
        """

    def submit_pipeline(
        self,
        deployment: "PipelineDeploymentResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            deployment: The pipeline deployment to submit.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment. These don't need to be set if running locally.
            placeholder_run: An optional placeholder run for the deployment.

        Returns:
            Optional submission result.
        """
        return None

    def prepare_or_run_pipeline(
        self,
        deployment: "PipelineDeploymentResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[Iterator[Dict[str, MetadataType]]]:
        """DEPRECATED: Prepare or run a pipeline.

        Args:
            deployment: The pipeline deployment to prepare or run.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment. These don't need to be set if running locally.
            placeholder_run: An optional placeholder run for the deployment.
        """

    def run(
        self,
        deployment: "PipelineDeploymentResponse",
        stack: "Stack",
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> None:
        """Runs a pipeline on a stack.

        Args:
            deployment: The pipeline deployment.
            stack: The stack on which to run the pipeline.
            placeholder_run: An optional placeholder run for the deployment.
                This will be deleted in case the pipeline deployment failed.

        Raises:
            KeyboardInterrupt: If the orchestrator is synchronous and the
                pipeline run is keyboard interrupted.
            RunMonitoringError: If a failure happened while monitoring the
                pipeline run.
        """
        self._prepare_run(deployment=deployment)

        pipeline_run_id: Optional[UUID] = None
        schedule_id: Optional[UUID] = None
        if deployment.schedule:
            schedule_id = deployment.schedule.id
        if placeholder_run:
            pipeline_run_id = placeholder_run.id

        environment = get_config_environment_vars(
            schedule_id=schedule_id,
            pipeline_run_id=pipeline_run_id,
        )

        prevent_client_side_caching = handle_bool_env_var(
            ENV_ZENML_PREVENT_CLIENT_SIDE_CACHING, default=False
        )

        if (
            placeholder_run
            and self.config.supports_client_side_caching
            and not deployment.schedule
            and not prevent_client_side_caching
        ):
            from zenml.orchestrators import cache_utils

            run_required = (
                cache_utils.create_cached_step_runs_and_prune_deployment(
                    deployment=deployment,
                    pipeline_run=placeholder_run,
                    stack=stack,
                )
            )

            if not run_required:
                self._cleanup_run()
                return
        else:
            logger.debug("Skipping client-side caching.")

        try:
            if (
                getattr(self.submit_pipeline, "__func__", None)
                is BaseOrchestrator.submit_pipeline
            ):
                logger.warning(
                    "The orchestrator '%s' is still using the deprecated "
                    "`prepare_or_run_pipeline(...)` method which will be "
                    "removed in the future. Please implement the replacement "
                    "`submit_pipeline(...)` method for your custom "
                    "orchestrator.",
                    self.name,
                )
                if metadata_iterator := self.prepare_or_run_pipeline(
                    deployment=deployment,
                    stack=stack,
                    environment=environment,
                    placeholder_run=placeholder_run,
                ):
                    for metadata_dict in metadata_iterator:
                        try:
                            if placeholder_run:
                                publish_pipeline_run_metadata(
                                    pipeline_run_id=placeholder_run.id,
                                    pipeline_run_metadata={
                                        self.id: metadata_dict
                                    },
                                )
                        except Exception as e:
                            logger.debug(
                                "Something went went wrong trying to publish the"
                                f"run metadata: {e}"
                            )
            else:
                submission_result = self.submit_pipeline(
                    deployment=deployment,
                    stack=stack,
                    environment=environment,
                    placeholder_run=placeholder_run,
                )

                if submission_result:
                    if submission_result.metadata:
                        if placeholder_run:
                            try:
                                publish_pipeline_run_metadata(
                                    pipeline_run_id=placeholder_run.id,
                                    pipeline_run_metadata={
                                        self.id: submission_result.metadata
                                    },
                                )
                            except Exception as e:
                                logger.debug(
                                    "Error publishing run metadata: %s", e
                                )
                        elif deployment.schedule:
                            try:
                                publish_schedule_metadata(
                                    schedule_id=deployment.schedule.id,
                                    schedule_metadata={
                                        self.id: submission_result.metadata
                                    },
                                )
                            except Exception as e:
                                logger.debug(
                                    "Error publishing schedule metadata: %s", e
                                )

                    if submission_result.wait_for_completion:
                        try:
                            submission_result.wait_for_completion()
                        except KeyboardInterrupt:
                            error_message = "Received KeyboardInterrupt. Note that the run is still executing. "
                            if placeholder_run:
                                error_message += (
                                    "If you want to stop the pipeline run, please use: "
                                    f"`zenml pipeline runs stop {placeholder_run.id}`"
                                )
                            raise KeyboardInterrupt(error_message)
                        except BaseException as e:
                            raise RunMonitoringError(original_exception=e)

        finally:
            self._cleanup_run()

    def run_step(self, step: "Step") -> None:
        """Runs the given step.

        Args:
            step: The step to run.

        Raises:
            RunStoppedException: If the run was stopped.
            BaseException: If the step failed all retries.
        """

        def _launch_step() -> None:
            assert self._active_deployment

            launcher = StepLauncher(
                deployment=self._active_deployment,
                step=step,
                orchestrator_run_id=self.get_orchestrator_run_id(),
            )
            launcher.launch()

        if self.config.handles_step_retries:
            _launch_step()
        else:
            # The orchestrator subclass doesn't handle step retries, so we
            # handle it in-process instead
            retries = 0
            retry_config = step.config.retry
            max_retries = retry_config.max_retries if retry_config else 0
            delay = retry_config.delay if retry_config else 0
            backoff = retry_config.backoff if retry_config else 1

            while retries <= max_retries:
                try:
                    _launch_step()
                except RunStoppedException:
                    # Don't retry if the run was stopped
                    raise
                except BaseException:
                    retries += 1
                    if retries <= max_retries:
                        logger.info(
                            "Sleeping for %d seconds before retrying step `%s`.",
                            delay,
                            step.config.name,
                        )
                        time.sleep(delay)
                        delay *= backoff
                    else:
                        if max_retries > 0:
                            logger.error(
                                "Failed to run step `%s` after %d retries.",
                                step.config.name,
                                max_retries,
                            )
                        raise
                else:
                    break

    @staticmethod
    def requires_resources_in_orchestration_environment(
        step: "Step",
    ) -> bool:
        """Checks if the orchestrator should run this step on special resources.

        Args:
            step: The step that will be checked.

        Returns:
            True if the step requires special resources in the orchestration
            environment, False otherwise.
        """
        # If the step requires custom resources and doesn't run with a step
        # operator, it would need these requirements in the orchestrator
        # environment
        if step.config.step_operator:
            return False

        return not step.config.resource_settings.empty

    def _prepare_run(self, deployment: "PipelineDeploymentResponse") -> None:
        """Prepares a run.

        Args:
            deployment: The deployment to prepare.
        """
        self._active_deployment = deployment

    def _cleanup_run(self) -> None:
        """Cleans up the active run."""
        self._active_deployment = None

    def fetch_status(
        self, run: "PipelineRunResponse", include_steps: bool = False
    ) -> Tuple[
        Optional[ExecutionStatus], Optional[Dict[str, ExecutionStatus]]
    ]:
        """Refreshes the status of a specific pipeline run.

        Args:
            run: A pipeline run response to fetch its status.
            include_steps: If True, also fetch the status of individual steps.

        Raises:
            NotImplementedError: If any orchestrator inheriting from the base
                class does not implement this logic.
        """
        raise NotImplementedError(
            "The fetch status functionality is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )

    def stop_run(
        self, run: "PipelineRunResponse", graceful: bool = False
    ) -> None:
        """Stops a specific pipeline run.

        This method should only be called if the orchestrator's
        supports_cancellation property is True.

        Args:
            run: A pipeline run response to stop.
            graceful: If True, allows for graceful shutdown where possible.
                If False, forces immediate termination. Default is False.

        Raises:
            NotImplementedError: If any orchestrator inheriting from the base
                class does not implement this logic.
        """
        # Check if the orchestrator supports cancellation
        if (
            getattr(self._stop_run, "__func__", None)
            is BaseOrchestrator._stop_run
        ):
            raise NotImplementedError(
                f"The '{self.__class__.__name__}' orchestrator does not "
                "support stopping pipeline runs."
            )

        # Update pipeline status to STOPPING before calling concrete implementation
        publish_pipeline_run_status_update(
            pipeline_run_id=run.id,
            status=ExecutionStatus.STOPPING,
        )

        # Now call the concrete implementation
        self._stop_run(run=run, graceful=graceful)

    def _stop_run(
        self, run: "PipelineRunResponse", graceful: bool = False
    ) -> None:
        """Concrete implementation of pipeline stopping logic.

        This method should be implemented by concrete orchestrator classes
        instead of stop_run to ensure proper status management.

        Args:
            run: A pipeline run response to stop (already updated to STOPPING status).
            graceful: If True, allows for graceful shutdown where possible.
                If False, forces immediate termination. Default is True.

        Raises:
            NotImplementedError: If any orchestrator inheriting from the base
                class does not implement this logic.
        """
        raise NotImplementedError(
            "The stop run functionality is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )

    @property
    def supports_schedule_updates(self) -> bool:
        """Whether the orchestrator supports updating schedules.

        Returns:
            Whether the orchestrator supports updating schedules.
        """
        return (
            getattr(self.update_schedule, "__func__", None)
            is not BaseOrchestrator.update_schedule
        )

    @property
    def supports_schedule_deletion(self) -> bool:
        """Whether the orchestrator supports deleting schedules.

        Returns:
            Whether the orchestrator supports deleting schedules.
        """
        return (
            getattr(self.delete_schedule, "__func__", None)
            is not BaseOrchestrator.delete_schedule
        )

    def update_schedule(
        self, schedule: "ScheduleResponse", update: "ScheduleUpdate"
    ) -> None:
        """Updates a schedule.

        Args:
            schedule: The schedule to update.
            update: The update to apply to the schedule.

        Raises:
            NotImplementedError: If the functionality is not implemented.
        """
        raise NotImplementedError(
            "Schedule updating is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )

    def delete_schedule(self, schedule: "ScheduleResponse") -> None:
        """Deletes a schedule.

        Args:
            schedule: The schedule to delete.

        Raises:
            NotImplementedError: If the functionality is not implemented.
        """
        raise NotImplementedError(
            "Schedule deletion is not implemented for the "
            f"'{self.__class__.__name__}' orchestrator."
        )


class BaseOrchestratorFlavor(Flavor):
    """Base orchestrator flavor class."""

    @property
    def type(self) -> StackComponentType:
        """Returns the flavor type.

        Returns:
            The flavor type.
        """
        return StackComponentType.ORCHESTRATOR

    @property
    def config_class(self) -> Type[BaseOrchestratorConfig]:
        """Config class for the base orchestrator flavor.

        Returns:
            The config class.
        """
        return BaseOrchestratorConfig

    @property
    @abstractmethod
    def implementation_class(self) -> Type["BaseOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of pipeline run tables."""

import json
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence
from uuid import UUID

from pydantic import ConfigDict
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import selectinload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import TEXT, Column, Field, Relationship

from zenml.config.pipeline_configurations import PipelineConfiguration
from zenml.config.step_configurations import Step
from zenml.constants import TEXT_FIELD_MAX_LENGTH
from zenml.enums import (
    ExecutionStatus,
    MetadataResourceTypes,
    TaggableResourceTypes,
)
from zenml.logger import get_logger
from zenml.models import (
    PipelineRunRequest,
    PipelineRunResponse,
    PipelineRunResponseBody,
    PipelineRunResponseMetadata,
    PipelineRunUpdate,
    RunMetadataEntry,
)
from zenml.models.v2.core.pipeline_run import PipelineRunResponseResources
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.constants import MODEL_VERSION_TABLENAME
from zenml.zen_stores.schemas.pipeline_build_schemas import PipelineBuildSchema
from zenml.zen_stores.schemas.pipeline_deployment_schemas import (
    PipelineDeploymentSchema,
)
from zenml.zen_stores.schemas.pipeline_schemas import PipelineSchema
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schedule_schema import ScheduleSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
)
from zenml.zen_stores.schemas.stack_schemas import StackSchema
from zenml.zen_stores.schemas.trigger_schemas import TriggerExecutionSchema
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import (
    RunMetadataInterface,
    jl_arg,
)

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.logs_schemas import LogsSchema
    from zenml.zen_stores.schemas.model_schemas import (
        ModelVersionPipelineRunSchema,
        ModelVersionSchema,
    )
    from zenml.zen_stores.schemas.run_metadata_schemas import RunMetadataSchema
    from zenml.zen_stores.schemas.service_schemas import ServiceSchema
    from zenml.zen_stores.schemas.step_run_schemas import StepRunSchema
    from zenml.zen_stores.schemas.tag_schemas import TagSchema

logger = get_logger(__name__)


class PipelineRunSchema(NamedSchema, RunMetadataInterface, table=True):
    """SQL Model for pipeline runs."""

    __tablename__ = "pipeline_run"
    __table_args__ = (
        UniqueConstraint(
            "deployment_id",
            "orchestrator_run_id",
            name="unique_orchestrator_run_id_for_deployment_id",
        ),
        UniqueConstraint(
            "name",
            "project_id",
            name="unique_run_name_in_project",
        ),
    )

    # Fields
    orchestrator_run_id: Optional[str] = Field(nullable=True)
    start_time: Optional[datetime] = Field(nullable=True)
    end_time: Optional[datetime] = Field(nullable=True, default=None)
    status: str = Field(nullable=False)
    orchestrator_environment: Optional[str] = Field(
        sa_column=Column(TEXT, nullable=True)
    )

    # Foreign keys
    deployment_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineDeploymentSchema.__tablename__,
        source_column="deployment_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=True,
    )
    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    pipeline_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineSchema.__tablename__,
        source_column="pipeline_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    model_version_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=MODEL_VERSION_TABLENAME,
        source_column="model_version_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    # Relationships
    deployment: Optional["PipelineDeploymentSchema"] = Relationship(
        back_populates="pipeline_runs"
    )
    project: "ProjectSchema" = Relationship(back_populates="runs")
    user: Optional["UserSchema"] = Relationship(back_populates="runs")
    run_metadata: List["RunMetadataSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            secondary="run_metadata_resource",
            primaryjoin=f"and_(foreign(RunMetadataResourceSchema.resource_type)=='{MetadataResourceTypes.PIPELINE_RUN.value}', foreign(RunMetadataResourceSchema.resource_id)==PipelineRunSchema.id)",
            secondaryjoin="RunMetadataSchema.id==foreign(RunMetadataResourceSchema.run_metadata_id)",
            overlaps="run_metadata",
        ),
    )
    logs: List["LogsSchema"] = Relationship(
        back_populates="pipeline_run",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    step_runs: List["StepRunSchema"] = Relationship(
        sa_relationship_kwargs={"cascade": "delete"},
    )
    model_version: "ModelVersionSchema" = Relationship(
        back_populates="pipeline_runs",
    )

    # Temporary fields and foreign keys to be deprecated
    pipeline_configuration: Optional[str] = Field(
        sa_column=Column(TEXT, nullable=True)
    )
    client_environment: Optional[str] = Field(
        sa_column=Column(TEXT, nullable=True)
    )

    stack_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=StackSchema.__tablename__,
        source_column="stack_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    build_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineBuildSchema.__tablename__,
        source_column="build_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    schedule_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=ScheduleSchema.__tablename__,
        source_column="schedule_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    trigger_execution_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=TriggerExecutionSchema.__tablename__,
        source_column="trigger_execution_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    stack: Optional["StackSchema"] = Relationship()
    build: Optional["PipelineBuildSchema"] = Relationship()
    schedule: Optional["ScheduleSchema"] = Relationship()
    pipeline: Optional["PipelineSchema"] = Relationship()
    trigger_execution: Optional["TriggerExecutionSchema"] = Relationship()

    services: List["ServiceSchema"] = Relationship(
        back_populates="pipeline_run",
    )
    tags: List["TagSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=f"and_(foreign(TagResourceSchema.resource_type)=='{TaggableResourceTypes.PIPELINE_RUN.value}', foreign(TagResourceSchema.resource_id)==PipelineRunSchema.id)",
            secondary="tag_resource",
            secondaryjoin="TagSchema.id == foreign(TagResourceSchema.tag_id)",
            order_by="TagSchema.name",
            overlaps="tags",
        ),
    )

    # Needed for cascade deletion
    model_versions_pipeline_runs_links: List[
        "ModelVersionPipelineRunSchema"
    ] = Relationship(
        back_populates="pipeline_run",
        sa_relationship_kwargs={"cascade": "delete"},
    )

    model_config = ConfigDict(protected_namespaces=())  # type: ignore[assignment]

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        from zenml.zen_stores.schemas import ModelVersionSchema

        options = [
            selectinload(jl_arg(PipelineRunSchema.deployment)).joinedload(
                jl_arg(PipelineDeploymentSchema.pipeline)
            ),
            selectinload(jl_arg(PipelineRunSchema.deployment)).joinedload(
                jl_arg(PipelineDeploymentSchema.stack)
            ),
            selectinload(jl_arg(PipelineRunSchema.deployment)).joinedload(
                jl_arg(PipelineDeploymentSchema.build)
            ),
            selectinload(jl_arg(PipelineRunSchema.deployment)).joinedload(
                jl_arg(PipelineDeploymentSchema.schedule)
            ),
            selectinload(jl_arg(PipelineRunSchema.deployment)).joinedload(
                jl_arg(PipelineDeploymentSchema.code_reference)
            ),
        ]

        # if include_metadata:
        #     options.extend(
        #         [
        #             joinedload(jl_arg(PipelineRunSchema.run_metadata)),
        #         ]
        #     )

        if include_resources:
            options.extend(
                [
                    selectinload(
                        jl_arg(PipelineRunSchema.model_version)
                    ).joinedload(
                        jl_arg(ModelVersionSchema.model), innerjoin=True
                    ),
                    selectinload(jl_arg(PipelineRunSchema.logs)),
                    selectinload(jl_arg(PipelineRunSchema.user)),
                    selectinload(jl_arg(PipelineRunSchema.tags)),
                ]
            )

        return options

    @classmethod
    def from_request(
        cls, request: "PipelineRunRequest"
    ) -> "PipelineRunSchema":
        """Convert a `PipelineRunRequest` to a `PipelineRunSchema`.

        Args:
            request: The request to convert.

        Returns:
            The created `PipelineRunSchema`.
        """
        orchestrator_environment = json.dumps(request.orchestrator_environment)
        if len(orchestrator_environment) > TEXT_FIELD_MAX_LENGTH:
            logger.warning(
                "Orchestrator environment is too large to be stored in the "
                "database. Skipping."
            )
            orchestrator_environment = "{}"

        return cls(
            project_id=request.project,
            user_id=request.user,
            name=request.name,
            orchestrator_run_id=request.orchestrator_run_id,
            orchestrator_environment=orchestrator_environment,
            start_time=request.start_time,
            status=request.status.value,
            pipeline_id=request.pipeline,
            deployment_id=request.deployment,
            trigger_execution_id=request.trigger_execution_id,
        )

    def get_pipeline_configuration(self) -> PipelineConfiguration:
        """Get the pipeline configuration for the pipeline run.

        Raises:
            RuntimeError: if the pipeline run has no deployment and no pipeline
                configuration.

        Returns:
            The pipeline configuration.
        """
        if self.deployment:
            pipeline_config = PipelineConfiguration.model_validate_json(
                self.deployment.pipeline_configuration
            )
        elif self.pipeline_configuration:
            pipeline_config = PipelineConfiguration.model_validate_json(
                self.pipeline_configuration
            )
        else:
            raise RuntimeError(
                "Pipeline run has no deployment and no pipeline configuration."
            )

        pipeline_config.finalize_substitutions(
            start_time=self.start_time, inplace=True
        )
        return pipeline_config

    def get_step_configuration(self, step_name: str) -> Step:
        """Get the step configuration for the pipeline run.

        Args:
            step_name: The name of the step to get the configuration for.

        Raises:
            RuntimeError: If the pipeline run has no deployment.

        Returns:
            The step configuration.
        """
        if self.deployment:
            pipeline_configuration = self.get_pipeline_configuration()
            return Step.from_dict(
                data=json.loads(
                    self.deployment.get_step_configuration(step_name).config
                ),
                pipeline_configuration=pipeline_configuration,
            )
        else:
            raise RuntimeError("Pipeline run has no deployment.")

    def fetch_metadata_collection(
        self, include_full_metadata: bool = False, **kwargs: Any
    ) -> Dict[str, List[RunMetadataEntry]]:
        """Fetches all the metadata entries related to the pipeline run.

        Args:
            include_full_metadata: Whether the full metadata will be included.
            **kwargs: Keyword arguments.

        Returns:
            a dictionary, where the key is the key of the metadata entry
                and the values represent the list of entries with this key.
        """
        # Fetch the metadata related to this run
        metadata_collection = super().fetch_metadata_collection(**kwargs)

        if include_full_metadata:
            # Fetch the metadata related to the steps of this run
            for s in self.step_runs:
                step_metadata = s.fetch_metadata_collection()
                for k, v in step_metadata.items():
                    metadata_collection[f"{s.name}::{k}"] = v

            # Fetch the metadata related to the schedule of this run
            if self.deployment is not None:
                if schedule := self.deployment.schedule:
                    schedule_metadata = schedule.fetch_metadata_collection()
                    for k, v in schedule_metadata.items():
                        metadata_collection[f"schedule:{k}"] = v

        return metadata_collection

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        include_python_packages: bool = False,
        include_full_metadata: bool = False,
        **kwargs: Any,
    ) -> "PipelineRunResponse":
        """Convert a `PipelineRunSchema` to a `PipelineRunResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            include_python_packages: Whether the python packages will be filled.
            include_full_metadata: Whether the full metadata will be included.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The created `PipelineRunResponse`.

        Raises:
            RuntimeError: if the model creation fails.
        """
        if self.deployment is not None:
            config = PipelineConfiguration.model_validate_json(
                self.deployment.pipeline_configuration
            )
            client_environment = json.loads(self.deployment.client_environment)

            stack = (
                self.deployment.stack.to_model()
                if self.deployment.stack
                else None
            )
            pipeline = (
                self.deployment.pipeline.to_model()
                if self.deployment.pipeline
                else None
            )
            build = (
                self.deployment.build.to_model()
                if self.deployment.build
                else None
            )
            schedule = (
                self.deployment.schedule.to_model()
                if self.deployment.schedule
                else None
            )
            code_reference = (
                self.deployment.code_reference.to_model()
                if self.deployment.code_reference
                else None
            )

        elif self.pipeline_configuration is not None:
            config = PipelineConfiguration.model_validate_json(
                self.pipeline_configuration
            )
            client_environment = (
                json.loads(self.client_environment)
                if self.client_environment
                else {}
            )

            stack = self.stack.to_model() if self.stack else None
            pipeline = self.pipeline.to_model() if self.pipeline else None
            build = self.build.to_model() if self.build else None
            schedule = self.schedule.to_model() if self.schedule else None
            code_reference = None

        else:
            raise RuntimeError(
                "Pipeline run model creation has failed. Each pipeline run "
                "entry should either have a deployment_id or "
                "pipeline_configuration."
            )

        config.finalize_substitutions(start_time=self.start_time, inplace=True)

        body = PipelineRunResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            status=ExecutionStatus(self.status),
            stack=stack,
            pipeline=pipeline,
            build=build,
            schedule=schedule,
            code_reference=code_reference,
            trigger_execution=(
                self.trigger_execution.to_model()
                if self.trigger_execution
                else None
            ),
            created=self.created,
            updated=self.updated,
            deployment_id=self.deployment_id,
            model_version_id=self.model_version_id,
        )
        metadata = None
        if include_metadata:
            is_templatable = False
            if (
                self.deployment
                and self.deployment.build
                and not self.deployment.build.is_local
                and self.deployment.build.stack_id
            ):
                is_templatable = True

            orchestrator_environment = (
                json.loads(self.orchestrator_environment)
                if self.orchestrator_environment
                else {}
            )

            if not include_python_packages:
                client_environment.pop("python_packages", None)
                orchestrator_environment.pop("python_packages", None)

            metadata = PipelineRunResponseMetadata(
                run_metadata=self.fetch_metadata(
                    include_full_metadata=include_full_metadata
                ),
                config=config,
                start_time=self.start_time,
                end_time=self.end_time,
                client_environment=client_environment,
                orchestrator_environment=orchestrator_environment,
                orchestrator_run_id=self.orchestrator_run_id,
                code_path=self.deployment.code_path
                if self.deployment
                else None,
                template_id=self.deployment.template_id
                if self.deployment
                else None,
                is_templatable=is_templatable,
            )

        resources = None
        if include_resources:
            # Add the client logs as "logs" if they exist, for backwards compatibility
            # TODO: This will be safe to remove in future releases (>0.84.0).
            client_logs = [
                log_entry
                for log_entry in self.logs
                if log_entry.source == "client"
            ]

            resources = PipelineRunResponseResources(
                user=self.user.to_model() if self.user else None,
                model_version=self.model_version.to_model()
                if self.model_version
                else None,
                tags=[tag.to_model() for tag in self.tags],
                logs=client_logs[0].to_model() if client_logs else None,
                log_collection=[log.to_model() for log in self.logs],
            )

        return PipelineRunResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(self, run_update: "PipelineRunUpdate") -> "PipelineRunSchema":
        """Update a `PipelineRunSchema` with a `PipelineRunUpdate`.

        Args:
            run_update: The `PipelineRunUpdate` to update with.

        Returns:
            The updated `PipelineRunSchema`.
        """
        if run_update.status:
            self.status = run_update.status.value
            self.end_time = run_update.end_time

        self.updated = utc_now()
        return self

    def update_placeholder(
        self, request: "PipelineRunRequest"
    ) -> "PipelineRunSchema":
        """Update a placeholder run.

        Args:
            request: The pipeline run request which should replace the
                placeholder.

        Raises:
            RuntimeError: If the DB entry does not represent a placeholder run.
            ValueError: If the run request is not a valid request to replace the
                placeholder run.

        Returns:
            The updated `PipelineRunSchema`.
        """
        if not self.is_placeholder_run():
            raise RuntimeError(
                f"Unable to replace pipeline run {self.id} which is not a "
                "placeholder run."
            )

        if request.is_placeholder_request:
            raise ValueError(
                "Cannot replace a placeholder run with another placeholder run."
            )

        if (
            self.deployment_id != request.deployment
            or self.pipeline_id != request.pipeline
            or self.project_id != request.project
        ):
            raise ValueError(
                "Deployment, project or pipeline ID of placeholder run "
                "do not match the IDs of the run request."
            )

        if not request.orchestrator_run_id:
            raise ValueError(
                "Orchestrator run ID is required to replace a placeholder run."
            )

        if (
            self.orchestrator_run_id
            and self.orchestrator_run_id != request.orchestrator_run_id
        ):
            raise ValueError(
                "Orchestrator run ID of placeholder run does not match the "
                "ID of the run request."
            )

        orchestrator_environment = json.dumps(request.orchestrator_environment)

        self.orchestrator_run_id = request.orchestrator_run_id
        self.orchestrator_environment = orchestrator_environment
        self.status = request.status.value

        self.updated = utc_now()

        return self

    def is_placeholder_run(self) -> bool:
        """Whether the pipeline run is a placeholder run.

        Returns:
            Whether the pipeline run is a placeholder run.
        """
        return self.status == ExecutionStatus.INITIALIZING.value



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Notebook utilities."""

import hashlib
from typing import Any, Callable, Optional, TypeVar, Union

from zenml.environment import Environment
from zenml.logger import get_logger

ZENML_NOTEBOOK_CELL_CODE_ATTRIBUTE_NAME = "__zenml_notebook_cell_code__"

AnyObject = TypeVar("AnyObject", bound=Any)

logger = get_logger(__name__)


def is_defined_in_notebook_cell(obj: Any) -> bool:
    """Check whether an object is defined in a notebook cell.

    Args:
        obj: The object to check.

    Returns:
        Whether the object is defined in a notebook cell.
    """
    if not Environment.in_notebook():
        return False

    module_name = getattr(obj, "__module__", None)
    return module_name == "__main__"


def enable_notebook_code_extraction(
    _obj: Optional["AnyObject"] = None,
) -> Union["AnyObject", Callable[["AnyObject"], "AnyObject"]]:
    """Decorator to enable code extraction from notebooks.

    Args:
        _obj: The class or function for which to enable code extraction.

    Returns:
        The decorated class or function.
    """

    def inner_decorator(obj: "AnyObject") -> "AnyObject":
        try_to_save_notebook_cell_code(obj)
        return obj

    if _obj is None:
        return inner_decorator
    else:
        return inner_decorator(_obj)


def get_active_notebook_cell_code() -> Optional[str]:
    """Get the code of the currently active notebook cell.

    Returns:
        The code of the currently active notebook cell.
    """
    cell_code = None
    try:
        ipython = get_ipython()  # type: ignore[name-defined]
        cell_code = ipython.get_parent()["content"]["code"]
    except (NameError, KeyError) as e:
        logger.warning("Unable to extract cell code: %s.", str(e))

    return cell_code


def try_to_save_notebook_cell_code(obj: Any) -> None:
    """Try to save the notebook cell code for an object.

    Args:
        obj: The object for which to save the notebook cell code.
    """
    if is_defined_in_notebook_cell(obj):
        if cell_code := get_active_notebook_cell_code():
            setattr(
                obj,
                ZENML_NOTEBOOK_CELL_CODE_ATTRIBUTE_NAME,
                cell_code,
            )


def load_notebook_cell_code(obj: Any) -> Optional[str]:
    """Load the notebook cell code for an object.

    Args:
        obj: The object for which to load the cell code.

    Returns:
        The notebook cell code if it was saved.
    """
    return getattr(obj, ZENML_NOTEBOOK_CELL_CODE_ATTRIBUTE_NAME, None)


def warn_about_notebook_cell_magic_commands(cell_code: str) -> None:
    """Warn about magic commands in the cell code.

    Args:
        cell_code: The cell code.
    """
    if any(line.startswith(("%", "!")) for line in cell_code.splitlines()):
        logger.warning(
            "Some lines in your notebook cell start with a `!` or `%` "
            "character. Running a ZenML step remotely from a notebook "
            "only works if the cell only contains python code. If any "
            "of these lines contain Jupyter notebook magic commands, "
            "remove them and try again."
        )


def compute_cell_replacement_module_name(cell_code: str) -> str:
    """Compute the replacement module name for a given cell code.

    Args:
        cell_code: The code of the notebook cell.

    Returns:
        The replacement module name.
    """
    code_hash = hashlib.sha1(cell_code.encode()).hexdigest()  # nosec
    return f"extracted_notebook_code_{code_hash}"



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes serialization utils."""

import re
from datetime import date, datetime
from typing import Any, Dict, List, Type, cast


def serialize_kubernetes_model(model: Any) -> Dict[str, Any]:
    """Serializes a Kubernetes model.

    Args:
        model: The model to serialize.

    Raises:
        TypeError: If the model is not a Kubernetes model.

    Returns:
        The serialized model.
    """
    if not is_model_class(model.__class__.__name__):
        raise TypeError(f"Unable to serialize non-kubernetes model {model}.")

    assert hasattr(model, "attribute_map")
    attribute_mapping = cast(Dict[str, str], model.attribute_map)

    model_attributes = {
        serialized_attribute_name: getattr(model, attribute_name)
        for attribute_name, serialized_attribute_name in attribute_mapping.items()
    }
    return _serialize_dict(model_attributes)


def _serialize_dict(dict_: Dict[str, Any]) -> Dict[str, Any]:
    """Serializes a dictionary.

    Args:
        dict_: Dict to serialize.

    Returns:
        The serialized dict.
    """
    return {key: _serialize(value) for key, value in dict_.items()}


def _serialize(value: Any) -> Any:
    """Serializes any valid object.

    Args:
        value: The object to serialize.

    Raises:
        TypeError: If the object is not serializable.

    Returns:
        The serialized object.
    """
    primitive_types = (float, bool, bytes, str, int)

    if value is None:
        return None
    elif isinstance(value, primitive_types):
        return value
    elif isinstance(value, (datetime, date)):
        return value.isoformat()
    elif isinstance(value, List):
        return [_serialize(item) for item in value]
    elif isinstance(value, tuple):
        return tuple(_serialize(item) for item in value)
    elif isinstance(value, Dict):
        return _serialize_dict(value)
    elif is_model_class(value.__class__.__name__):
        return serialize_kubernetes_model(value)
    else:
        raise TypeError(f"Failed to serialize unknown object {value}")


def deserialize_kubernetes_model(data: Dict[str, Any], class_name: str) -> Any:
    """Deserializes a Kubernetes model.

    Args:
        data: The model data.
        class_name: Name of the Kubernetes model class.

    Raises:
        KeyError: If the data contains values for an invalid attribute.

    Returns:
        The deserialized model.
    """
    model_class = get_model_class(class_name=class_name)
    assert hasattr(model_class, "openapi_types")
    assert hasattr(model_class, "attribute_map")
    # Mapping of the attribute name of the model class to the attribute type
    type_mapping = cast(Dict[str, str], model_class.openapi_types)
    reverse_attribute_mapping = cast(Dict[str, str], model_class.attribute_map)
    # Mapping of the serialized key to the attribute name of the model class
    attribute_mapping = {
        value: key for key, value in reverse_attribute_mapping.items()
    }

    deserialized_attributes: Dict[str, Any] = {}

    for key, value in data.items():
        if key not in attribute_mapping:
            raise KeyError(
                f"Got value for attribute {key} which is not one of the "
                f"available attributes for class {class_name}: "
                f"{set(attribute_mapping)}."
            )

        attribute_name = attribute_mapping[key]
        attribute_class = type_mapping[attribute_name]

        if not value:
            deserialized_attributes[attribute_name] = value
        elif attribute_class.startswith("list["):
            match = re.fullmatch(r"list\[(.*)\]", attribute_class)
            assert match
            inner_class = match.group(1)
            deserialized_attributes[attribute_name] = _deserialize_list(
                value, class_name=inner_class
            )
        elif attribute_class.startswith("dict("):
            match = re.fullmatch(r"dict\(([^,]*), (.*)\)", attribute_class)
            assert match
            inner_class = match.group(1)
            deserialized_attributes[attribute_name] = _deserialize_dict(
                value, class_name=inner_class
            )
        elif is_model_class(attribute_class):
            deserialized_attributes[attribute_name] = (
                deserialize_kubernetes_model(value, attribute_class)
            )
        else:
            deserialized_attributes[attribute_name] = value

    return model_class(**deserialized_attributes)


def is_model_class(class_name: str) -> bool:
    """Checks whether the given class name is a Kubernetes model class.

    Args:
        class_name: Name of the class to check.

    Returns:
        If the given class name is a Kubernetes model class.
    """
    import kubernetes.client.models

    return hasattr(kubernetes.client.models, class_name)


def get_model_class(class_name: str) -> Type[Any]:
    """Gets a Kubernetes model class.

    Args:
        class_name: Name of the class to get.

    Raises:
        TypeError: If no Kubernetes model class exists for this name.

    Returns:
        The model class.
    """
    import kubernetes.client.models

    class_ = getattr(kubernetes.client.models, class_name, None)

    if not class_:
        raise TypeError(
            f"Unable to find kubernetes model class with name {class_name}."
        )

    assert isinstance(class_, type)
    return class_


def _deserialize_list(data: Any, class_name: str) -> List[Any]:
    """Deserializes a list of potential Kubernetes models.

    Args:
        data: The data to deserialize.
        class_name: Name of the class of the elements of the list.

    Returns:
        The deserialized list.
    """
    assert isinstance(data, List)
    if is_model_class(class_name):
        return [
            deserialize_kubernetes_model(element, class_name)
            for element in data
        ]
    else:
        return data


def _deserialize_dict(data: Any, class_name: str) -> Dict[str, Any]:
    """Deserializes a dict of potential Kubernetes models.

    Args:
        data: The data to deserialize.
        class_name: Name of the class of the elements of the dict.

    Returns:
        The deserialized dict.
    """
    assert isinstance(data, Dict)
    if is_model_class(class_name):
        return {
            key: deserialize_kubernetes_model(value, class_name)
            for key, value in data.items()
        }
    else:
        return data



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base implementation of actions."""

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, Optional, Type

from zenml.enums import PluginType
from zenml.event_hub.base_event_hub import BaseEventHub
from zenml.logger import get_logger
from zenml.models import (
    ActionFlavorResponse,
    ActionFlavorResponseBody,
    ActionFlavorResponseMetadata,
    ActionRequest,
    ActionResponse,
    ActionUpdate,
    TriggerExecutionResponse,
)
from zenml.models.v2.base.base import BaseResponse
from zenml.plugins.base_plugin_flavor import (
    BasePlugin,
    BasePluginConfig,
    BasePluginFlavor,
)
from zenml.zen_server.auth import AuthContext
from zenml.zen_server.rbac.models import ResourceType

logger = get_logger(__name__)


# -------------------- Configuration Models ----------------------------------


class ActionConfig(BasePluginConfig):
    """Allows configuring the action configuration."""


# -------------------- Flavors ---------------------------------------------


class BaseActionFlavor(BasePluginFlavor, ABC):
    """Base Action Flavor to register Action Configurations."""

    TYPE: ClassVar[PluginType] = PluginType.ACTION

    # Action specific
    ACTION_CONFIG_CLASS: ClassVar[Type[ActionConfig]]

    @classmethod
    def get_action_config_schema(cls) -> Dict[str, Any]:
        """The config schema for a flavor.

        Returns:
            The config schema.
        """
        return cls.ACTION_CONFIG_CLASS.model_json_schema()

    @classmethod
    def get_flavor_response_model(cls, hydrate: bool) -> ActionFlavorResponse:
        """Convert the Flavor into a Response Model.

        Args:
            hydrate: Whether the model should be hydrated.

        Returns:
            The response model.
        """
        metadata = None
        if hydrate:
            metadata = ActionFlavorResponseMetadata(
                config_schema=cls.get_action_config_schema(),
            )
        return ActionFlavorResponse(
            body=ActionFlavorResponseBody(),
            metadata=metadata,
            name=cls.FLAVOR,
            type=cls.TYPE,
            subtype=cls.SUBTYPE,
        )


# -------------------- Plugin -----------------------------------


class BaseActionHandler(BasePlugin, ABC):
    """Implementation for an action handler."""

    _event_hub: Optional[BaseEventHub] = None

    def __init__(self, event_hub: Optional[BaseEventHub] = None) -> None:
        """Event source handler initialization.

        Args:
            event_hub: Optional event hub to use to initialize the event source
                handler. If not set during initialization, it may be set
                at a later time by calling `set_event_hub`. An event hub must
                be configured before the event handler needs to dispatch events.
        """
        super().__init__()
        if event_hub is None:
            from zenml.event_hub.event_hub import (
                event_hub as default_event_hub,
            )

            # TODO: for now, we use the default internal event hub. In
            # the future, this should be configurable.
            event_hub = default_event_hub

        self.set_event_hub(event_hub)

    @property
    @abstractmethod
    def config_class(self) -> Type[ActionConfig]:
        """Returns the `BasePluginConfig` config.

        Returns:
            The configuration.
        """

    @property
    @abstractmethod
    def flavor_class(self) -> Type[BaseActionFlavor]:
        """Returns the flavor class of the plugin.

        Returns:
            The flavor class of the plugin.
        """

    @property
    def event_hub(self) -> BaseEventHub:
        """Get the event hub used to dispatch events.

        Returns:
            The event hub.

        Raises:
            RuntimeError: if an event hub isn't configured.
        """
        if self._event_hub is None:
            raise RuntimeError(
                f"An event hub is not configured for the "
                f"{self.flavor_class.FLAVOR} {self.flavor_class.SUBTYPE} "
                f"action handler."
            )

        return self._event_hub

    def set_event_hub(self, event_hub: BaseEventHub) -> None:
        """Configure an event hub for this event source plugin.

        Args:
            event_hub: Event hub to be used by this event handler to dispatch
                events.
        """
        self._event_hub = event_hub
        self._event_hub.subscribe_action_handler(
            action_flavor=self.flavor_class.FLAVOR,
            action_subtype=self.flavor_class.SUBTYPE,
            callback=self.event_hub_callback,
        )

    def event_hub_callback(
        self,
        config: Dict[str, Any],
        trigger_execution: TriggerExecutionResponse,
        auth_context: AuthContext,
    ) -> None:
        """Callback to be used by the event hub to dispatch events to the action handler.

        Args:
            config: The action configuration
            trigger_execution: The trigger execution
            auth_context: Authentication context with an API token that can be
                used by external workloads to authenticate with the server
                during the execution of the action. This API token is associated
                with the service account that was configured for the trigger
                that activated the action and has a validity defined by the
                trigger's authentication window.
        """
        try:
            config_obj = self.config_class(**config)
        except ValueError as e:
            logger.exception(
                f"Action handler {self.flavor_class.FLAVOR} of type "
                f"{self.flavor_class.SUBTYPE} received an invalid configuration "
                f"from the event hub: {e}."
            )
            return

        try:
            # TODO: this would be a great place to convert the event back into its
            # original form and pass it to the action handler.
            self.run(
                config=config_obj,
                trigger_execution=trigger_execution,
                auth_context=auth_context,
            )
        except Exception:
            # Don't let the event hub crash if the action handler fails
            # TODO: we might want to return a value here indicating to the event
            # hub that the action handler failed to execute the action. This can
            # be used by the event hub to retry the action handler or to log the
            # failure.
            logger.exception(
                f"Action handler {self.flavor_class.FLAVOR} of type "
                f"{self.flavor_class.SUBTYPE} failed to execute the action "
                f"with configuration {config}."
            )

    @abstractmethod
    def run(
        self,
        config: ActionConfig,
        trigger_execution: TriggerExecutionResponse,
        auth_context: AuthContext,
    ) -> None:
        """Execute an action.

        Args:
            config: The action configuration
            trigger_execution: The trigger execution
            auth_context: Authentication context with an API token that can be
                used by external workloads to authenticate with the server
                during the execution of the action. This API token is associated
                with the service account that was configured for the trigger
                that activated the action and has a validity defined by the
                trigger's authentication window.
        """

    def create_action(self, action: ActionRequest) -> ActionResponse:
        """Process a action request and create the action in the database.

        Args:
            action: Action request.

        Raises:
            Exception: If the implementation specific processing before creating
                the action fails.

        Returns:
            The created action.
        """
        # Validate and instantiate the configuration from the request
        config = self.validate_action_configuration(action.configuration)
        # Call the implementation specific method to validate the request
        # before it is sent to the database
        self._validate_action_request(action=action, config=config)
        # Serialize the configuration back into the request
        action.configuration = config.model_dump(exclude_none=True)
        # Create the action in the database
        action_response = self.zen_store.create_action(action=action)
        try:
            # Instantiate the configuration from the response
            config = self.validate_action_configuration(action.configuration)
            # Call the implementation specific method to process the created
            # action
            self._process_action_request(action=action_response, config=config)
            # Add any implementation specific related resources to the action
            # response
            self._populate_action_response_resources(
                action=action_response,
                config=config,
            )
        except Exception:
            # If the action creation fails, delete the action from
            # the database
            logger.exception(
                f"Failed to create action {action_response}. "
                f"Deleting the action."
            )
            self.zen_store.delete_action(action_id=action_response.id)
            raise

        # Serialize the configuration back into the response
        action_response.set_configuration(config.model_dump(exclude_none=True))
        # Return the response to the user
        return action_response

    def update_action(
        self,
        action: ActionResponse,
        action_update: ActionUpdate,
    ) -> ActionResponse:
        """Process action update and update the action in the database.

        Args:
            action: The action to update.
            action_update: The update to be applied to the action.

        Raises:
            Exception: If the implementation specific processing before updating
                the action fails.

        Returns:
            The updated action.
        """
        # Validate and instantiate the configuration from the original event
        # source
        config = self.validate_action_configuration(action.configuration)
        # Validate and instantiate the configuration from the update request
        # NOTE: if supplied, the configuration update is a full replacement
        # of the original configuration
        config_update = config
        if action_update.configuration is not None:
            config_update = self.validate_action_configuration(
                action_update.configuration
            )
        # Call the implementation specific method to validate the update request
        # before it is sent to the database
        self._validate_action_update(
            action=action,
            config=config,
            action_update=action_update,
            config_update=config_update,
        )
        # Serialize the configuration update back into the update request
        action_update.configuration = config_update.model_dump(
            exclude_none=True
        )

        # Update the action in the database
        action_response = self.zen_store.update_action(
            action_id=action.id,
            action_update=action_update,
        )
        try:
            # Instantiate the configuration from the response
            response_config = self.validate_action_configuration(
                action_response.configuration
            )
            # Call the implementation specific method to process the update
            # request before it is sent to the database
            self._process_action_update(
                action=action_response,
                config=response_config,
                previous_action=action,
                previous_config=config,
            )
            # Add any implementation specific related resources to the action
            # response
            self._populate_action_response_resources(
                action=action_response, config=response_config
            )
        except Exception:
            # If the action update fails, roll back the action in
            # the database to the original state
            logger.exception(
                f"Failed to update action {action}. "
                f"Rolling back the action to the previous state."
            )
            self.zen_store.update_action(
                action_id=action.id,
                action_update=ActionUpdate.from_response(action),
            )
            raise

        # Serialize the configuration back into the response
        action_response.set_configuration(
            response_config.model_dump(exclude_none=True)
        )
        # Return the response to the user
        return action_response

    def delete_action(
        self,
        action: ActionResponse,
        force: bool = False,
    ) -> None:
        """Process action delete request and delete the action in the database.

        Args:
            action: The action to delete.
            force: Whether to force delete the action from the database
                even if the action handler fails to delete the event
                source.

        Raises:
            Exception: If the implementation specific processing before deleting
                the action fails.
        """
        # Validate and instantiate the configuration from the original event
        # source
        config = self.validate_action_configuration(action.configuration)
        try:
            # Call the implementation specific method to process the deleted
            # action before it is deleted from the database
            self._process_action_delete(
                action=action,
                config=config,
                force=force,
            )
        except Exception:
            logger.exception(f"Failed to delete action {action}. ")
            if not force:
                raise

            logger.warning(f"Force deleting action {action}.")

        # Delete the action from the database
        self.zen_store.delete_action(
            action_id=action.id,
        )

    def get_action(
        self, action: ActionResponse, hydrate: bool = False
    ) -> ActionResponse:
        """Process a action response before it is returned to the user.

        Args:
            action: The action fetched from the database.
            hydrate: Whether to hydrate the action.

        Returns:
            The action.
        """
        if hydrate:
            # Instantiate the configuration from the response
            config = self.validate_action_configuration(action.configuration)
            # Call the implementation specific method to process the response
            self._process_action_response(action=action, config=config)
            # Serialize the configuration back into the response
            action.set_configuration(config.model_dump(exclude_none=True))
            # Add any implementation specific related resources to the action
            # response
            self._populate_action_response_resources(
                action=action, config=config
            )

        # Return the response to the user
        return action

    def validate_action_configuration(
        self, action_config: Dict[str, Any]
    ) -> ActionConfig:
        """Validate and return the action configuration.

        Args:
            action_config: The action configuration to validate.

        Returns:
            The validated action configuration.

        Raises:
            ValueError: if the configuration is invalid.
        """
        try:
            return self.config_class(**action_config)
        except ValueError as e:
            raise ValueError(f"Invalid configuration for action: {e}.") from e

    def extract_resources(
        self,
        action_config: ActionConfig,
        hydrate: bool = False,
    ) -> Dict[ResourceType, BaseResponse[Any, Any, Any]]:
        """Extract related resources for this action.

        Args:
            action_config: Action configuration from which to extract related
                resources.
            hydrate: Whether to hydrate the resource models.

        Returns:
            List of resources related to the action.
        """
        return {}

    def _validate_action_request(
        self, action: ActionRequest, config: ActionConfig
    ) -> None:
        """Validate an action request before it is created in the database.

        Concrete action handlers should override this method to add
        implementation specific functionality pertaining to the validation of
        a new action. The implementation may also modify the action
        request and/or configuration in place to apply implementation specific
        changes before the request is stored in the database.

        If validation is required, the implementation should raise a
        ValueError if the configuration is invalid. If any exceptions are raised
        during the validation, the action will not be created in the
        database.

        The resulted action configuration is serialized back into the action
        request before it is stored in the database.

        The implementation should not use this method to provision any external
        resources, as the action may not be created in the database if
        the database level validation fails.

        Args:
            action: Action request.
            config: Action configuration instantiated from the request.
        """
        pass

    def _process_action_request(
        self, action: ActionResponse, config: ActionConfig
    ) -> None:
        """Process an action request after it is created in the database.

        Concrete action handlers should override this method to add
        implementation specific functionality pertaining to the creation of
        a new action. The implementation may also modify the action
        response and/or configuration in place to apply implementation specific
        changes before the response is returned to the user.

        The resulted configuration is serialized back into the action
        response before it is returned to the user.

        The implementation should use this method to provision any external
        resources required for the action. If any of the provisioning
        fails, the implementation should raise an exception and the action
        will be deleted from the database.

        Args:
            action: Newly created action
            config: Action configuration instantiated from the response.
        """
        pass

    def _validate_action_update(
        self,
        action: ActionResponse,
        config: ActionConfig,
        action_update: ActionUpdate,
        config_update: ActionConfig,
    ) -> None:
        """Validate an action update before it is reflected in the database.

        Concrete action handlers should override this method to add
        implementation specific functionality pertaining to validation of an
        action update request. The implementation may also modify the
        action update and/or configuration update in place to apply
        implementation specific changes.

        If validation is required, the implementation should raise a
        ValueError if the configuration update is invalid. If any exceptions are
        raised during the validation, the action will not be updated in
        the database.

        The resulted configuration update is serialized back into the event
        source update before it is stored in the database.

        The implementation should not use this method to provision any external
        resources, as the action may not be updated in the database if
        the database level validation fails.

        Args:
            action: Original action before the update.
            config: Action configuration instantiated from the original
                trigger.
            action_update: Action update request.
            config_update: Action configuration instantiated from the
                updated action.
        """
        pass

    def _process_action_update(
        self,
        action: ActionResponse,
        config: ActionConfig,
        previous_action: ActionResponse,
        previous_config: ActionConfig,
    ) -> None:
        """Process an action after it is updated in the database.

        Concrete action handlers should override this method to add
        implementation specific functionality pertaining to updating an existing
        action. The implementation may also modify the action
        and/or configuration in place to apply implementation specific
        changes before the response is returned to the user.

        The resulted configuration is serialized back into the action
        response before it is returned to the user.

        The implementation should use this method to provision any external
        resources required for the action update. If any of the
        provisioning fails, the implementation should raise an exception and the
        action will be rolled back to the previous state in the database.

        Args:
            action: Action after the update.
            config: Action configuration instantiated from the updated
                action.
            previous_action: Original action before the update.
            previous_config: Action configuration instantiated from the
                original action.
        """
        pass

    def _process_action_delete(
        self,
        action: ActionResponse,
        config: ActionConfig,
        force: Optional[bool] = False,
    ) -> None:
        """Process an action before it is deleted from the database.

        Concrete action handlers should override this method to add
        implementation specific functionality pertaining to the deletion of an
        action.

        The implementation should use this method to deprovision any external
        resources required for the action. If any of the deprovisioning
        fails, the implementation should raise an exception and the
        action will kept in the database (unless force is True, in which
        case the action will be deleted from the database regardless of
        any deprovisioning failures).

        Args:
            action: Action before the deletion.
            config: Action configuration before the deletion.
            force: Whether to force deprovision the action.
        """
        pass

    def _process_action_response(
        self, action: ActionResponse, config: ActionConfig
    ) -> None:
        """Process an action response before it is returned to the user.

        Concrete action handlers should override this method to add
        implementation specific functionality pertaining to how the action
        response is returned to the user. The implementation may modify the
        the action response and/or configuration in place.

        The resulted configuration is serialized back into the action
        response before it is returned to the user.

        This method is applied to all action responses fetched from the
        database before they are returned to the user, with the exception of
        those returned from the `create_action`, `update_action`,
        and `delete_action` methods, which have their own specific
        processing methods.

        Args:
            action: Action response.
            config: Action configuration instantiated from the response.
        """
        pass

    def _populate_action_response_resources(
        self,
        action: ActionResponse,
        config: ActionConfig,
        hydrate: bool = False,
    ) -> None:
        """Populate related resources for the action response.

        Concrete action handlers should override this method to add
        implementation specific related resources to the action response.

        This method is applied to all action responses fetched from the
        database before they are returned to the user, including
        those returned from the `create_action`, `update_action`,
        and `delete_action` methods.

        Args:
            action: Action response.
            config: Action configuration instantiated from the response.
            hydrate: Whether to hydrate the resource models.
        """
        if action.resources is None:
            # We only populate the resources if the resources field is already
            # set by the database by means of hydration.
            return
        extract_resources = self.extract_resources(config, hydrate=hydrate)
        for resource_type, resource in extract_resources.items():
            setattr(action.resources, str(resource_type), resource)



"""Release [0.84.0].

Revision ID: 0.84.0
Revises: 360fa84718bf
Create Date: 2025-07-10 14:07:30.176139

"""

# revision identifiers, used by Alembic.
revision = "0.84.0"
down_revision = "360fa84718bf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



"""Release [0.74.0].

Revision ID: 0.74.0
Revises: 4d5524b92a30
Create Date: 2025-02-06 08:10:51.286515

"""

# revision identifiers, used by Alembic.
revision = "0.74.0"
down_revision = "4d5524b92a30"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Constants for the VertexAI integration."""

from google.cloud.aiplatform_v1.types.job_state import JobState

VERTEX_ENDPOINT_SUFFIX = "-aiplatform.googleapis.com"
POLLING_INTERVAL_IN_SECONDS = 10
CONNECTION_ERROR_RETRY_LIMIT = 5
_VERTEX_JOB_STATE_SUCCEEDED = JobState.JOB_STATE_SUCCEEDED
_VERTEX_JOB_STATE_FAILED = JobState.JOB_STATE_FAILED
_VERTEX_JOB_STATE_CANCELLED = JobState.JOB_STATE_CANCELLED

VERTEX_JOB_STATES_COMPLETED = (
    _VERTEX_JOB_STATE_SUCCEEDED,
    _VERTEX_JOB_STATE_FAILED,
    _VERTEX_JOB_STATE_CANCELLED,
)
VERTEX_JOB_STATES_FAILED = (
    _VERTEX_JOB_STATE_FAILED,
    _VERTEX_JOB_STATE_CANCELLED,
)

GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL = (
    "cloud.google.com/gke-accelerator"
)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Amazon S3 artifact store flavor."""

import re
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Optional,
    Set,
    Type,
)

from pydantic import Field, field_validator

from zenml.artifact_stores import (
    BaseArtifactStoreConfig,
    BaseArtifactStoreFlavor,
)
from zenml.integrations.s3 import S3_ARTIFACT_STORE_FLAVOR
from zenml.integrations.s3.utils import split_s3_path
from zenml.models import ServiceConnectorRequirements
from zenml.stack.authentication_mixin import AuthenticationConfigMixin
from zenml.utils.networking_utils import (
    replace_localhost_with_internal_hostname,
)
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.s3.artifact_stores import S3ArtifactStore


class S3ArtifactStoreConfig(
    BaseArtifactStoreConfig, AuthenticationConfigMixin
):
    """Configuration for the S3 Artifact Store.

    All attributes of this class except `path` will be passed to the
    `s3fs.S3FileSystem` initialization. See
    [here](https://s3fs.readthedocs.io/en/latest/) for more information on how
    to use those configuration options to connect to any S3-compatible storage.

    When you want to register an S3ArtifactStore from the CLI and need to pass
    `client_kwargs`, `config_kwargs` or `s3_additional_kwargs`, you should pass
    them as a json string:
    ```
    zenml artifact-store register my_s3_store --flavor=s3 \
    --path=s3://my_bucket --client_kwargs='{"endpoint_url": "http://my-s3-endpoint"}'
    ```
    """

    SUPPORTED_SCHEMES: ClassVar[Set[str]] = {"s3://"}

    key: Optional[str] = SecretField(
        default=None,
        description="AWS access key ID for authentication. "
        "If not provided, credentials will be inferred from the environment.",
    )
    secret: Optional[str] = SecretField(
        default=None,
        description="AWS secret access key for authentication. "
        "If not provided, credentials will be inferred from the environment.",
    )
    token: Optional[str] = SecretField(
        default=None,
        description="AWS session token for temporary credentials. "
        "If not provided, credentials will be inferred from the environment.",
    )
    client_kwargs: Optional[Dict[str, Any]] = Field(
        None,
        description="Additional keyword arguments to pass to the S3 client. "
        "For example, to connect to a custom S3-compatible endpoint: "
        "{'endpoint_url': 'http://minio:9000'}",
    )
    config_kwargs: Optional[Dict[str, Any]] = Field(
        None,
        description="Additional keyword arguments to pass to the S3 client configuration. "
        "For example: {'region_name': 'us-west-2', 'signature_version': 's3v4'}",
    )
    s3_additional_kwargs: Optional[Dict[str, Any]] = Field(
        None,
        description="Additional keyword arguments for S3 operations. "
        "For example: {'ACL': 'bucket-owner-full-control'}",
    )

    _bucket: Optional[str] = None

    @field_validator("client_kwargs")
    @classmethod
    def _validate_client_kwargs(
        cls, value: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        """Validates the `client_kwargs` attribute.

        Args:
            value: The value to validate.

        Raises:
            ValueError: If the value is not a valid URL.

        Returns:
            The validated value.
        """
        if value is None:
            return value

        if "endpoint_url" in value and value["endpoint_url"]:
            url = value["endpoint_url"].rstrip("/")
            scheme = re.search("^([a-z0-9]+://)", url)
            if scheme is None or scheme.group() not in ("https://", "http://"):
                raise ValueError(
                    "Invalid URL for endpoint url: {url}. Should be in the form "
                    "https://hostname[:port] or http://hostname[:port]."
                )

            # When running inside a container, if the URL uses localhost, the
            # target service will not be available. We try to replace localhost
            # with one of the special Docker or K3D internal hostnames.
            value["endpoint_url"] = replace_localhost_with_internal_hostname(
                url
            )
        return value

    @property
    def bucket(self) -> str:
        """The bucket name of the artifact store.

        Returns:
            The bucket name of the artifact store.
        """
        if self._bucket is None:
            self._bucket, _ = split_s3_path(self.path)
        return self._bucket


class S3ArtifactStoreFlavor(BaseArtifactStoreFlavor):
    """Flavor of the S3 artifact store."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return S3_ARTIFACT_STORE_FLAVOR

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            resource_type="s3-bucket",
            resource_id_attr="path",
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A URL to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A URL to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A URL to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/artifact_store/aws.png"

    @property
    def config_class(self) -> Type[S3ArtifactStoreConfig]:
        """The config class of the flavor.

        Returns:
            The config class of the flavor.
        """
        return S3ArtifactStoreConfig

    @property
    def implementation_class(self) -> Type["S3ArtifactStore"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class for this flavor.
        """
        from zenml.integrations.s3.artifact_stores import S3ArtifactStore

        return S3ArtifactStore
